
/* simple-cling.cc - compile and execute string - using cling Interpreter */

#include "cling/Interpreter/Interpreter.h"

#include "clang/Frontend/HeaderSearchOptions.h"
#include "clang/Frontend/CompilerInstance.h"
#include "clang/Serialization/ASTReader.h"

#include "TSystem.h"
#include "TString.h"
#include "TObjArray.h"
#include "TObjString.h"

#include <iostream>

//______________________________________________________________________________
void addPath(clang::CompilerInstance* CI, const TString& path)
{
   // std::cout << "addPath " << path << std::endl;
   clang::HeaderSearchOptions& headerOpts = CI->getHeaderSearchOpts();
   const bool IsUserSupplied = false;
   const bool IsFramework = false;
   const bool IsSysRootRelative = false;
   headerOpts.AddPath(path.Data(), clang::frontend::Angled, IsUserSupplied, IsFramework, IsSysRootRelative);
}

//______________________________________________________________________________
void addIncludePath(clang::CompilerInstance* CI)
{
   TString inclPath = gSystem->GetIncludePath();
   // std::cout << "IncludePath " << inclPath << std::endl;
   TObjArray* list = inclPath.Tokenize(" ");

   int cnt = list->GetEntries();
   for (int i = 0; i < cnt ; i++) {
      TObjString* p = dynamic_cast <TObjString*>(list->At(i));
      TString s = p->GetString();

      if (s.BeginsWith("-I"))
         s = s.Remove(0, 2);

      if (s.BeginsWith("\"") && s.EndsWith("\"") && s.Length() != 1) {
         s = s.Remove(0, 1);
         s = s.Remove(s.Length() - 1, 1);
      }

      // s = s.Strip (TString::kBoth);

      // std::cout << "IncludeItem " << s << std::endl;

      if (s != "") {
         gSystem->ExpandPathName(s);
         addPath(CI, s);
      }
   }

   delete list;
}

//______________________________________________________________________________
void defineSymbol(clang::PreprocessorOptions& P, const TString& name, const TString& value)
{
   // std::cout << "defineSymbol " << name << "=" << value << "!" << std::endl;

   TString s = name;
   if (value != "")
      s = s + "=" + value;

   P.addMacroUndef(name.Data());
   P.addMacroDef(s.Data());
}

//______________________________________________________________________________
void gccOptions(clang::CompilerInstance* CI, TString gcc = "")
{
   if (gcc == "")
      gcc = "gcc";

   TString cmd = "echo 'int main () { }' | " + gcc + " -v -x c++ -dD -E - 2>&1";
   TString txt = gSystem->GetFromPipe(cmd);

   // include directories

   {
      TString pattern = "#include <...> search starts here:";
      Ssiz_t area_start = txt.Index(pattern);
      if (area_start != kNPOS)
         area_start = area_start + pattern.Length();

      Ssiz_t area_stop = txt.Index("End of search list");

      if (area_start != kNPOS && area_stop != kNPOS) {
         Ssiz_t inx = area_start;
         Ssiz_t len = area_stop;

         while (inx < len) {
            while (inx < len && txt [inx] <= ' ') inx++;
            int token_start = inx;

            while (inx < len && txt [inx] > ' ') inx++;
            int token_stop = inx;

            TString dir = txt(token_start, token_stop - token_start);
            if (dir != "") {
               addPath(CI, dir);  // system directory
            }
         }
      }
   }

   // conditional symbols

   {
      clang::PreprocessorOptions& P = CI->getPreprocessorOpts();

      Ssiz_t inx = 0;
      Ssiz_t total = txt.Length();

      while (inx < total) {
         // find end of line
         Ssiz_t lim = inx;
         while (lim < total && (txt [lim] >= ' ' || txt [lim] == '\t')) lim ++;
         // std::cout << "line: " << txt (inx, lim-inx) << std::endl;

         // one word
         while (inx < lim && txt [inx] <= ' ') inx++;
         int token_start = inx;
         while (inx < lim && txt [inx] > ' ') inx++;
         TString key = txt(token_start, inx - token_start);

         if (key == "#define") {
            // symbol name (only simple identifier)
            while (inx < lim && txt [inx] <= ' ') inx++;
            int token_start = inx;
            while (inx < lim && txt [inx] > ' ') inx++;
            TString name = txt(token_start, inx - token_start);

            while (inx < lim && txt [inx] <= ' ') inx++;
            TString value = txt(inx, lim - inx); // rest of line
            defineSymbol(P, name, value);
         }

         // next line
         while (lim < total && (txt [lim] < ' ' && txt [lim] != '\t')) lim ++;
         inx = lim;
      }
   }
}

//______________________________________________________________________________
void printInfo(clang::CompilerInstance* CI)
{
   const clang::HeaderSearchOptions& hsOpts = CI->getHeaderSearchOpts();
   std::cout << "Sysroot " << hsOpts.Sysroot << std::endl;
   int size = hsOpts.UserEntries.size();
   for (int i = 0; i < size; i++)
      std::cout << "User entry " << hsOpts.UserEntries[i].Path << std::endl;
   std::cout << "EnvIncPath " << hsOpts.EnvIncPath << std::endl;
   std::cout << "CEnvIncPath " << hsOpts.CEnvIncPath << std::endl;
   std::cout << "CXXEnvIncPath " << hsOpts.CXXEnvIncPath << std::endl;
   std::cout << "ResourceDir " << hsOpts.ResourceDir << std::endl;
}

//______________________________________________________________________________
cling::Interpreter* createInterpreter()
{
   cling::Interpreter* fInterpreter = new cling::Interpreter();
   clang::CompilerInstance* CI = fInterpreter->getCI();

   clang::LangOptions& langInfo = CI->getLangOpts();
   //langInfo.C99         = 1;
   //langInfo.HexFloats   = 1;
   langInfo.BCPLComment = 1; // Only for C99/C++.
   langInfo.Digraphs    = 1; // C94, C99, C++.
   langInfo.CPlusPlus   = 1;
   //langInfo.CPlusPlus0x = 1;
   langInfo.CXXOperatorNames = 1;
   langInfo.Bool = 1;
   langInfo.NeXTRuntime = 1;
   langInfo.NoInline = 1;
   //: langInfo.Exceptions = 1;
   langInfo.GNUMode = 1;
   langInfo.NoInline = 1;
   langInfo.GNUInline = 1;
   langInfo.DollarIdents = 1;
   //: langInfo.POSIXThreads = 1;

   #if 0
      CI->getHeaderSearchOpts().Verbose = 1;
   #endif

   CI->getHeaderSearchOpts().UseStandardIncludes = false;
   CI->getHeaderSearchOpts().UseStandardCXXIncludes = false;

   CI->getDiagnosticOpts().ShowColors = true;

   // addPath (CI, headerOpts.ResourceDir); // llvm/lib/clang/1.5

   gccOptions(CI);  // gcc include directories

   addIncludePath(CI);  // include/root

   // printInfo(CI);

   return fInterpreter;
}

#if 0
//______________________________________________________________________________
void clang2reflex(cling::Interpreter* p)
{
   clang::CompilerInstance* CI = p->getCI();

   clang::TranslationUnitDecl* tu =
      CI->getASTContext().getTranslationUnitDecl();

   ClrScan(& CI->getASTContext(), tu);

   info ("writing output/output.html");
   ClrPrint("output/output.txt");
   ClrPrint("output/output.html", "style.css", true);
}
#endif

//______________________________________________________________________________
void openPCH(clang::CompilerInstance* CI, const TString& name)
{
   std::string fileName = name.Data();

   clang::PreprocessorOptions & PO = CI->getInvocation().getPreprocessorOpts();

   std::string originalFile =
      #ifdef LLVM_2_8
      clang::ASTReader::getOriginalSourceFile(fileName, CI->getDiagnostics());
      #else
      clang::ASTReader::getOriginalSourceFile(fileName, CI->getFileManager(), CI->getFileSystemOpts(), CI->getDiagnostics());
      #endif

   if (! originalFile.empty())
   {
      PO.ImplicitPCHInclude = fileName;
      PO.Includes.push_back(originalFile);
      std::cout << "precompiled header: " + fileName << std::endl;
      std::cout << "original file name: " + originalFile << std::endl;
   }
   else std::cout << "Cannot find original file for precompiled header: " + fileName  << std::endl;
}


//______________________________________________________________________________
int main(int argc, const char **argv)
{
   cling::Interpreter* p = createInterpreter();

   clang::CompilerInstance* CI = p->getCI();

   #if 0
   CI->getInvocation().getPreprocessorOpts().ImplicitPCHInclude = "output/example4.pch";
   CI->getInvocation().getPreprocessorOpts().Includes.push_back("input/example4.cc");
   #endif

   openPCH (CI, "_output/example4.pch");

   p->processLine("int i = 1; \n");
   p->processLine("i++; \n");
   p->processLine("i = f3 (2, 3); \n");
   p->processLine("printf (\"Hello, i = %d \\n\", i); \n");

   // clang2reflex (p);

   // p->executeFile("/abc/gibbon/prog/mix/clr-try/input/test.cc"); // !? ??

   // p->processLine("void ClrTest ();");
   // p->processLine("ClrTest ();");

   // p->processLine ("#include <TGButton.h>");
   // p->processLine ("TGButton* b = new TGButton ();");

   // p->processLine("#include <TCanvas.h> \n");
   // p->processLine("TCanvas* c = new TCanvas;");

   // p->loadFile ("input/example3.cc");

   // ClrInfo ();

   return 0;
}

