
/* simple-detail.cc - compile and execute string - code copied from cling Interpreter */

#include "clang/AST/ASTConsumer.h"
#include "clang/AST/ASTContext.h"
#include "clang/AST/DeclGroup.h"
#include "clang/Basic/TargetInfo.h"
#include "clang/CodeGen/ModuleBuilder.h"
#include "clang/Frontend/ASTConsumers.h"
#include "clang/Frontend/CompilerInstance.h"
#include "clang/Serialization/ASTReader.h"
#include "clang/Frontend/TextDiagnosticBuffer.h"
#include "clang/Lex/Preprocessor.h"
#include "clang/Parse/ParseAST.h"

#include "llvm/LLVMContext.h"
#include "llvm/Module.h"
#include "llvm/PassManager.h"

#include "llvm/Analysis/Verifier.h"
#include "llvm/Assembly/PrintModulePass.h"
#include "llvm/ExecutionEngine/ExecutionEngine.h"
#include "llvm/ExecutionEngine/GenericValue.h"
#include "llvm/Support/MemoryBuffer.h"
#include "llvm/Support/ManagedStatic.h"
#include "llvm/Target/TargetSelect.h"

#include "llvm/System/Signals.h" // function PrintStackTraceOnErrorSignal

#include <stdio.h>
#include <iostream>
#include <stdexcept> // class std::runtime_error

using namespace std;

// use LLVM and CLang as shared libraries, otherwise execution engine is not created

/* ---------------------------------------------------------------------- */

llvm::LLVMContext * m_llvm_context = NULL;
clang::CompilerInstance * m_CI = NULL;
llvm::ExecutionEngine* m_engine = NULL;
bool m_printAST = false;

static const char* fake_argv [] = { "clang", "-x", "c++",
                                     // "-include-pch", "output/example40.pch",
                                     // "-include-pth", "output/tgbutton.pth",
                                     0 };
static const int fake_argc = (sizeof(fake_argv) / sizeof(const char*)) - 1;

/* ---------------------------------------------------------------------- */

void info (string msg)
{
   cout << msg << endl;
}

void error (string msg)
{
   cerr << msg << endl;
   throw new std::runtime_error (msg);
}

/* ---------------------------------------------------------------------- */

clang::ASTConsumer * maybeGenerateASTPrinter ()
{
   if (m_printAST) {
      return clang::CreateASTPrinter(&llvm::outs());
   }
   return new clang::ASTConsumer();
}

/* ---------------------------------------------------------------------- */

void openPCH (const string fileName)
{
   clang::PreprocessorOptions & PO = m_CI->getInvocation().getPreprocessorOpts();

   std::string originalFile =
      #ifdef LLVM_2_8
      clang::ASTReader::getOriginalSourceFile(fileName, m_CI->getDiagnostics());
      #else
      clang::ASTReader::getOriginalSourceFile(fileName, m_CI->getFileManager(), m_CI->getFileSystemOpts(), m_CI->getDiagnostics());
      #endif

   if (! originalFile.empty())
   {
      PO.ImplicitPCHInclude = fileName;
      PO.Includes.push_back(originalFile);
      std::cout << "precompiled header: " + fileName << std::endl;
      std::cout << "original file name: " + originalFile << std::endl;
   }
   else std::cout << "Cannot find original file for precompiled header: " + fileName  << std::endl;
}

/* ---------------------------------------------------------------------- */

void init ()
{
   llvm::InitializeAllTargets();
   llvm::InitializeAllAsmPrinters();

   m_llvm_context = new llvm::LLVMContext;

   /* create compiler instance */

   m_CI = new clang::CompilerInstance();
   m_CI->setLLVMContext(m_llvm_context);

   /* setup diagnostics */

   //  Buffer the error messages while we process the compiler options.
   clang::TextDiagnosticBuffer DiagsBuffer;
   clang::Diagnostic Diags (&DiagsBuffer);
   clang::CompilerInvocation::CreateFromArgs
             (m_CI->getInvocation(),
              fake_argv + 1, fake_argv + fake_argc, Diags);

   m_CI->createDiagnostics (fake_argc - 1, const_cast<char**>(fake_argv + 1));
   if (!m_CI->hasDiagnostics())
     error ("Cannot create diagnostics");

   // Output the buffered error messages now.
   DiagsBuffer.FlushDiagnostics(m_CI->getDiagnostics());
   if (m_CI->getDiagnostics().getNumErrors() != 0)
      error ("Bad compiler options");

   /* set target */

   m_CI->setTarget
   (
      clang::TargetInfo::CreateTargetInfo
         (m_CI->getDiagnostics (),
          m_CI->getTargetOpts ())
   );

   if (!m_CI->hasTarget ())
      error ("No target");

   m_CI->getTarget().setForcedLangOptions(m_CI->getLangOpts());
   m_CI->createFileManager();

   #ifdef LLVM_2_8
   m_CI->createSourceManager();
   #else
   m_CI->createSourceManager (m_CI->getFileManager(), m_CI->getFileSystemOpts());
   #endif

   #if 0
   m_CI->createDiagnostics (fake_argc - 1, const_cast<char**>(fake_argv + 1));
   if (! m_CI->hasDiagnostics ())
      error ("Cannot create diagnostics");
   #endif

   /* create engine */

   llvm::Module * m_prev_module = new llvm::Module("_Clang_first", *m_llvm_context);
   // Note: Engine takes ownership of the module.
   llvm::EngineBuilder builder(m_prev_module);

   #if 0
   std::string errMsg;
   builder.setErrorStr (&errMsg);
   #endif

   builder.setEngineKind (llvm::EngineKind::JIT);
   // builder.setEngineKind (llvm::EngineKind::Interpreter);
   // builder.setEngineKind (llvm::EngineKind::Either);

   m_engine = builder.create();
   if (m_engine == NULL)
      error ("Unable to create the execution engine");
}

/* ---------------------------------------------------------------------- */

void cleanup ()
{
   delete m_engine;
   m_engine = 0;

   m_CI->takeLLVMContext(); // Don't take down the context with the m_CI.
   delete m_CI;
   m_CI = NULL;

   delete m_llvm_context;
   m_llvm_context = NULL;

   llvm::llvm_shutdown ();
}

/* ---------------------------------------------------------------------- */

void printModule(llvm::Module* m)
{
   llvm::PassManager PM;
   PM.add (llvm::createPrintModulePass (& llvm::outs ()));
   PM.run (*m);
}

/* ---------------------------------------------------------------------- */

void processLine (string input_line)
{
   string src = "";
   src += "#include <stdio.h>\n";
   src += "void __cling_internal() {\n";
   src += input_line;
   src += "\n}";

   // cout << "input_line: " << src << endl;

   /* compile string */

   m_CI->createPreprocessor();
   clang::Preprocessor & PP = m_CI->getPreprocessor();

   m_CI->getDiagnosticClient ().BeginSourceFile (m_CI->getLangOpts(), &PP);

   m_CI->setASTContext
   (
      new clang::ASTContext
             (m_CI->getLangOpts (),
              PP.getSourceManager (),
              m_CI->getTarget (),
              PP.getIdentifierTable(),
              PP.getSelectorTable (),
              PP.getBuiltinInfo (),
              // false,
              0)
   );

   m_CI->setASTConsumer (maybeGenerateASTPrinter ());

   #if 1
   PP.getBuiltinInfo().InitializeBuiltins
      (PP.getIdentifierTable (),
       PP.getLangOptions().NoBuiltin);
   #endif

   llvm::MemoryBuffer * SB = llvm::MemoryBuffer::getMemBufferCopy(src, "ENTRY_BUFFER");
   if (SB == NULL)
      error ("Failed to create memory buffer");

   // m_CI->getSourceManager().clearIDTables();
   m_CI->getSourceManager().createMainFileIDForMemBuffer(SB);
   if (m_CI->getSourceManager().getMainFileID().isInvalid())
      error ("compileString: Failed to create main file id");

   clang::ParseAST (PP, & m_CI->getASTConsumer (), m_CI->getASTContext ());
   #if 0
   m_CI->setASTConsumer (NULL);
   if (m_CI->hasPreprocessor())
      m_CI->getPreprocessor().EndSourceFile();
   m_CI->clearOutputFiles(m_CI->getDiagnostics().getNumErrors());
   m_CI->getDiagnosticClient().EndSourceFile();
   #endif

   unsigned err_count = m_CI->getDiagnostics().getNumErrors();
   if (err_count != 0)
      error ("Parse failed");

   clang::TranslationUnitDecl* tu = m_CI->getASTContext().getTranslationUnitDecl();
   if (tu == NULL)
      error ("Parse failed, no translation unit");

   /* generate code */

   llvm::OwningPtr <clang::CodeGenerator> codeGen
   (
      CreateLLVMCodeGen (m_CI->getDiagnostics (),
                         "ENTRY_MODULE",
                         m_CI->getCodeGenOpts (),
                         m_CI->getLLVMContext ())
   );

   codeGen->Initialize (m_CI->getASTContext ());

   info ("Running code generation");
   clang::TranslationUnitDecl::decl_iterator iter = tu->decls_begin ();
   clang::TranslationUnitDecl::decl_iterator iter_end = tu->decls_end ();
   for ( ; iter != iter_end; ++ iter )
      codeGen->HandleTopLevelDecl (clang::DeclGroupRef (*iter));
   codeGen->HandleTranslationUnit(m_CI->getASTContext ());
   info ("Finished code generation");

   llvm::Module* m = codeGen->ReleaseModule();
   if (m == NULL)
      error ("Module creation failed");

   #if 0
   printModule (m);
   #endif

   m_engine->addModule(m); // Note: The engine takes ownership of the module.

   /* execute */

   info ("Running generated code");
   m_engine->runStaticConstructorsDestructors (false);

   llvm::Function* f = m_engine->FindFunctionNamed ("_Z16__cling_internalv");
   if (f == NULL)
      error ("Could not find the __cling_internal() function");

   std::vector<llvm::GenericValue> args;
   llvm::GenericValue ret = m_engine->runFunction (f, args);

   m_engine->runStaticConstructorsDestructors (true);
   m_engine->freeMachineCodeForFunction (f);

   info ("Finished running generated code");
}

/* ---------------------------------------------------------------------- */

int main (int argc, const char * *argv)
{
   #if 1
      llvm::sys::PrintStackTraceOnErrorSignal();
      llvm::PrettyStackTraceProgram X(argc, argv);
   #endif

   init ();

   #if 0
   m_CI->getInvocation().getPreprocessorOpts().ImplicitPCHInclude = "output/example4.pch";
   m_CI->getInvocation().getPreprocessorOpts().Includes.push_back("input/example4.cc");
   #endif

   openPCH ("output/_example4.pch");

   processLine ("int i = 1; i ++; i = f3 (1, 2); printf (\"Hello i = %d \\n\", i);");
   cleanup ();
   return 0;
}

